"""
ABCs for planning/orchestration.
"""

from abc import ABC, abstractmethod
from typing import Type

from vertebrate.compute import Command, Environment
from vertebrate.data import Address, Location, Namespace


class Order(Location):

    @property
    @abstractmethod
    def command(self) -> Command:
        pass

    def __call__(self) -> None:
        # Enforce returning a receipt?
        self.command()

    @property
    @abstractmethod
    def requirements(self) -> list[Address]:
        pass


class Orders(Namespace):

    @property
    @abstractmethod
    def LocationType(self) -> Type[Order]:
        pass

    @abstractmethod
    def environment(self, env: str) -> Environment:
        # Order namespaces are responsible for Environment resolution
        pass


# Should Plans be proper Environments? They are collections of executable Orders
class Plan(ABC):

    def __init__(self, orders: Orders) -> None:
        self._orders = orders

    @property
    def orders(self) -> Orders:
        return self._orders

    @abstractmethod
    def make(self, address: Address) -> None:
        pass



"""
An example set of executables and order specs.
"""


def populate_file(path: str, content: str) -> None:
    with open(path, "w") as file_out:
        file_out.write(content)


def read_file(path: str) -> None:
    with open(path, "r") as file_in:
        print(file_in.read())


specs = {
    "file1": {
        "env": "PythonEnvironment",
        "exe": "populate_file",
        "args": ["file1.txt", "Hello, Vertebrate"],
        "requirements": [],
    },
    "copy": {
        "env": "OSEnvironment",
        "exe": "cp",
        "args": ["file1.txt", "file2.txt"],
        "requirements": ["file1"],
    },
    "file2": {
        "env": "PythonEnvironment",
        "exe": "read_file",
        "args": ["file2.txt"],
        "requirements": ["copy"],
    }
}



"""
Implementation of a Plan that executes orders linearly.
"""

from collections.abc import Iterator
import logging

from vertebrate.builtins.dag import DAGPlan
from vertebrate.data import Address


logging.basicConfig(
    format="%(asctime)s %(levelname)s: %(message)s",
    level=logging.INFO,
    datefmt="%Y-%m-%d %H:%M:%S",
)


class LinearPlan(DAGPlan):

    def make(self, address: Address) -> None:
        for order_name in self._plan(address):
            logging.info(f"Satisfying {order_name}")
            self.orders[order_name]()



"""
Implementation of a Plan that uses the Dagster framework for orchestration.
"""

import re

import dagster as dg

from vertebrate.builtins.dag import DAGPlan
from vertebrate.builtins.utils import format_address
from vertebrate.data import Address
from vertebrate.plan import Orders


class DagsterPlan(DAGPlan):

    def _construct_asset(self, order_name: Address) -> dg.AssetsDefinition:

        ins = {
            format_address(req): dg.AssetIn(key=format_address(req))
            for req in self.orders[order_name].requirements
        }
        @dg.asset(name=format_address(order_name), ins=ins)
        def _asset(**kwargs):
            self.orders[order_name]()

        return _asset

    def assets(self, address: Address) -> None:
        return [
            self._construct_asset(order_name) for order_name in self._plan(address)
        ]

    def make(self, address: Address) -> None:
        dg.materialize(self.assets(address))



"""
Implementation of a PSQL execution environment.
"""

import os
import subprocess
from typing import Any

from vertebrate.compute import Environment


class PSQLEnvironment(Environment):
    """
    Execute psql invocations with connection info passed as arguments.
    """

    def __init__(self, user: str, dbname: str, host: str, port: str):
        self.user = user
        self.dbname = dbname
        self.host = host
        self.port = port
        self.psql = self._find_psql_on_path()

    def _find_psql_on_path(self):
        completed = subprocess.run(("which", "psql"), env=os.environ, capture_output=True)
        psql_path = completed.stdout.decode('utf-8').strip()
        if psql_path == '':
            raise ValueError("psql not found on PATH")
        return psql_path

    def _connection_args(self):
        return (
            "-U", self.user,
            "-d", self.dbname,
            "-h", self.host,
            "-p", self.port,
        )

    def execute(
        self,
        executable: str,
        args: tuple = (),
        kwargs: dict = {},
    ) -> Any:
        completed = subprocess.run(
            (self.psql,) + self._connection_args() + ("-f", executable),
            env=kwargs, capture_output=True
        )
        return completed.stdout.decode("utf-8")

    def __contains__(self, item: str) -> bool:
        return os.path.exists(item)



"""
Shared helper objects.
"""

import re
from typing import Any

from vertebrate.data import Address


def format_address(address: Address) -> str:
    return re.sub('[^a-zA-Z0-9]', '_', address).strip('_')


class ObjectByAddress(object):

    def __init__(self, dict_: dict[str, Any]):
        self.dict_ = dict_

    def __getitem__(self, address: Address) -> type:
        return self.dict_[format_address(address)]

    def __setitem__(self, address: Address, obj: Any) -> None:
        self.dict_[format_address(address)] = obj



"""
Example usage of the Plan base class using the DagsterPlan built-in.
"""

from vertebrate.builtins.dagster import DagsterPlan
from vertebrate.builtins.orders import SimpleOrders
from vertebrate.builtins.python import PythonEnvironment
from vertebrate.examples import orders as examples


def main():
    print("++ NOTE: This example requires the 'dagster' extra dependency group")
    print("-> pip install vertebrate[dagster]")
    print()

    print("++ Initialize some simple orders")
    print("-> orders = SimpleOrders(examples.specs, PythonEnvironment(examples))")
    orders = SimpleOrders(examples.specs, PythonEnvironment(examples))
    print()

    print("++ Initialize a Dagster-based execution plan")
    print("-> plan = Dagster(orders)")
    plan = DagsterPlan(orders)
    print()

    print("++ Fulfill the order named 'file2'")
    print('-> plan.make("file2")')
    print("NOTE: This creates file1.txt and file2.txt in your working directory")
    plan.make("file2")
    print()


if __name__ == "__main__":
    main()



"""
An implementation of a file namespace.
"""

from pathlib import Path
from typing import IO, Type

from vertebrate.data import Namespace, Location, Address


# Datatypes that will be inferred by file extension
CSV = "csv"
JSON = "json"
JSONLINES = "jsonl"
YAML = "yaml"
PARQUET = "parquet"
XML = "xml"
HTML = "html"
TOML = "toml"


class CSV: ...
class JSON: ...
class JSONLINES: ...
class YAML: ...
class PARQUET: ...
class XML: ...
class HTML: ...
class TOML: ...


class File(Location):

    def open(self, *args, **kwargs) -> IO:
        return open(self.address, *args, **kwargs)


class FileSystem(Namespace):

    extensions = {
        ".csv": CSV,
        ".json": JSON,
        ".jsonl": JSONLINES,
        ".jl": JSONLINES,
        ".yaml": YAML,
        ".yml": YAML,
        ".parquet": PARQUET,
        ".xml": XML,
        ".html": HTML,
        ".htm": HTML,
        ".toml": TOML,
    }

    def type(self, address: Address) -> type:
        ext = Path(address).suffix
        if ext in self.extensions:
            return self.extensions[ext]
        else:
            return bytes

    @property
    def LocationType(self) -> Type[Location]:
        return File

    def __contains__(self, address: Address) -> bool:
        return Path(address).exists()

    def __delitem__(self, address: Address) -> None:
        Path(address).unlink()



"""
Implementations of ABCs that use the operating system directly
(ex. processes, files, sockets, etc.).
"""

import shutil
import subprocess
from typing import Any

from vertebrate.compute import Environment


class OSEnvironment(Environment):
    """
    Execute processes on the local operating system.
    """

    def _exe_path(self, item: str) -> str:
        return shutil.which(item)

    def __contains__(self, item: str) -> bool:
        return self._exe_path(item) is not None

    def execute(self, executable: str, args: tuple = (), kwargs: dict = {}) -> Any:
        exe_path = self._exe_path(executable)
        completed = subprocess.run((exe_path,) + args, env=kwargs, capture_output=True)
        return completed.stdout



"""
Implementation of a Plan that uses the Luigi framework for orchestration.
"""

import luigi

from vertebrate.builtins.dag import DAGPlan
from vertebrate.builtins.utils import format_address, ObjectByAddress
from vertebrate.data import Address


tasks = ObjectByAddress(globals())


class LuigiPlan(DAGPlan):

    def _define_tasks(self, address: Address) -> None:

        orders = self.orders

        for order_name in self._plan(address):

            def run(self):
                orders[self.name]()
                with self.output().open("w") as file_out:
                    file_out.write("DONE")

            def requires(self):
                return [
                    tasks[req]()
                    for req in orders[self.name].requirements
                ]

            def output(self):
                # A good place to save receipts if/when implemented
                return luigi.LocalTarget(f"{format_address(self.name)}")

            namespace = {
                "__module__": __name__,
                "name": order_name,
                "run": run,
                "requires": requires,
                "output": output,
            }

            Task = type(format_address(order_name), (luigi.Task,), namespace)
            tasks[order_name] = Task

    def make(self, address: Address) -> None:
        self._define_tasks(address)
        luigi.build([tasks[address]()], local_scheduler=True)



"""
Base class for Plans that make use of Order DAGs.
"""

import collections as cl
from collections.abc import Iterator

import networkx as nx

from vertebrate.data import Address
from vertebrate.plan import Plan


type DAG = dict[Address, list[Address]]
type Edge = tuple[Address, Address]


class DAGPlan(Plan):

    def _populate_dag(self, dag: DAG, address: Address) -> None:
        order = self.orders[address]
        for req in order.requirements:
            dag[address].append(req)
            if req not in dag:
                self._populate_dag(dag, req)

    def dag(self, address: Address) -> DAG:
        dag = cl.defaultdict(list)
        self._populate_dag(dag, address)
        return dag

    def _dag_edges(self, dag: DAG) -> Iterator[Edge]:
        for to_, froms in dag.items():
            for from_ in froms:
                yield from_, to_

    def _plan(self, address: Address) -> Iterator[Address]:
        D = nx.DiGraph()
        D.add_edges_from(self._dag_edges(self.dag(address)))
        yield from nx.topological_sort(D)



"""
ABCs for computing environments and executables.
"""

from abc import ABC, abstractmethod
from typing import Any


# Should Environments be proper Namespaces? Some Locations can be executable
class Environment(ABC):

    @abstractmethod
    def execute(self, executable: str, args: tuple[Any], kwargs: dict[str, Any]) -> Any:
        pass

    @abstractmethod
    def __contains__(self, item: str) -> bool:
        pass

    def __getitem__(self, key: str) -> "Action":
        return Action(environment=self, executable=key)


class Action(object):

    def __init__(self, environment: Environment, executable: str) -> None:
        self._env = environment
        self._exe = executable

    @property
    def environment(self) -> Environment:
        return self._env

    @property
    def executable(self) -> str:
        return self._exe

    def __call__(self, *args, **kwargs) -> Any:
        return self.environment.execute(self.executable, args, kwargs)


class Command(object):

    def __init__(
        self,
        action: Action,
        args: tuple[Any, ...] = (),
        kwargs: dict[str, Any] = {}
    ) -> None:
        self._action = action
        self._args = args
        self._kwargs = kwargs

    @property
    def action(self) -> Action:
        return self._action

    @property
    def args(self) -> tuple[Any, ...]:
        return self._args

    @property
    def kwargs(self) -> dict[str, Any]:
        return self._kwargs

    def __call__(self) -> Any:
        return self.action(*self.args, **self.kwargs)



"""
Implement a catalog of transfers.
"""

import networkx as nx
from pandas import DataFrame
from sqlalchemy import Table

from vertebrate.builtins.files import File, CSV, JSON, PARQUET
from vertebrate.builtins.variables import Dictionary, Value
from vertebrate.data import (
    Namespace,
    Location,
    LocationTypes,
    TemporaryLocation,
    Transfers,
    TransferImplementation,
)


### Transfer implementations


def dict_to_json_file(source: Value, target: File) -> None:
    import json
    with target.open('w') as file_out:
        json.dump(source.get(), file_out)


def json_file_to_dict(source: File, target: Value) -> None:
    import json
    with source.open('r') as file_in:
        target.set(json.load(file_in))


def table_to_df(source: Table, target: Value) -> None:
    from pandas import read_sql
    target.set(read_sql(source.address, source.namespace.engine))


def df_to_table(source: Value, target: Table) -> None:
    source.get().to_sql(
        target.address,
        target.namespace.engine,
        index=False,
    )


def parquet_to_df(source: File, target: Value) -> None:
    from pandas import read_parquet
    target.set(read_parquet(source.address))


def df_to_parquet(source: Value, target: File) -> None:
    source.get().to_parquet(
        target.address,
        engine="pyarrow",
        index=False,
    )


### Transfer catalog


class ExplicitTransfers(Transfers):

    catalog = {
        (dict, JSON): dict_to_json_file,
        (JSON, dict): json_file_to_dict,
        (Table, DataFrame): table_to_df,
        (DataFrame, Table): df_to_table,
        (PARQUET, DataFrame): parquet_to_df,
        (DataFrame, PARQUET): df_to_parquet,
    }

    def __getitem__(self, datatypes: LocationTypes) -> TransferImplementation | None:
        return self.catalog.get(datatypes)

    def __contains__(self, datatypes: LocationTypes) -> bool:
        return datatypes in self.catalog


def transfer_graph(edge_list):
    DG = nx.DiGraph()
    DG.add_edges_from(edge_list)
    return DG


HOW = "how"
variables = Dictionary(globals())


class PathTransfers(Transfers):

    catalog = transfer_graph([
        (from_, to_, {HOW: how})
        for ((from_, to_), how) in ExplicitTransfers.catalog.items()
    ])

    # Register namespaces used for generating temporary locations
    namespaces = {
        dict: variables,
        DataFrame: variables,
    }

    def _path(self, datatypes: LocationTypes) -> list[type]:
        source_type, target_type = datatypes
        try:
            return nx.shortest_path(self.catalog, source=source_type, target=target_type)
        except nx.exception.NetworkXNoPath:
            pass

    def _temp_loc(self, loc_type: type) -> TemporaryLocation:
        return self.namespaces[loc_type].temp()

    def _exec(self, step: LocationTypes, src_loc: Location, tgt_loc: Location) -> None:
        execute_this_step = self.catalog.edges[*step][HOW]
        execute_this_step(src_loc, tgt_loc)
        if isinstance(src_loc, TemporaryLocation):
            del src_loc.namespace[src_loc.address]

    def _exec_intermediate(self, step: LocationTypes, src_loc: Location) -> Location:
        _, target_type = step
        tgt_loc = self._temp_loc(target_type)
        self._exec(step, src_loc, tgt_loc)
        return tgt_loc

    def __getitem__(self, datatypes: LocationTypes) -> TransferImplementation | None:
        path = self._path(datatypes)
        if path is not None:
            def transfer(source: Location, target: Location) -> None:
                *steps, last = nx.utils.pairwise(path)
                src_loc = source
                for step in steps:
                    src_loc = self._exec_intermediate(step, src_loc)
                self._exec(last, src_loc, target)
            return transfer

    def __contains__(self, datatypes: LocationTypes) -> bool:
        return self._path(datatypes) is not None



"""
Implementations of ABCs using pure Python.
"""

import sys
from typing import Any, Protocol, runtime_checkable

from vertebrate.compute import Environment


@runtime_checkable
class WithAttributes(Protocol):

    __dict__: dict[str, Any]


class PythonEnvironment(Environment):
    """
    Execute functions or methods defined on a Python object.
    """

    def __init__(self, obj: WithAttributes):
        self.obj = obj

    def __contains__(self, item: str) -> bool:
        return hasattr(self.obj, item)

    def execute(self, executable: str, args: tuple = (), kwargs: dict = {}) -> Any:
        return getattr(self.obj, executable)(*args, **kwargs)



"""
Example showing data transfer between Locations among Namespaces.
"""

from pandas import DataFrame

from vertebrate.builtins.files import FileSystem
from vertebrate.builtins.tables import Database
from vertebrate.builtins.transfers import ExplicitTransfers, PathTransfers
from vertebrate.builtins.variables import Dictionary
from vertebrate.data import Transfer


if __name__ == "__main__":
    print("++ NOTE: This example requires the 'transfers' extra dependency group")
    print("-> pip install vertebrate[transfers]")
    print()

    print("++ Define a catalog of ways to transfer data")
    print("-> Transfer.catalog = ExplicitTransfers()")
    Transfer.catalog = ExplicitTransfers()
    print()

    print("++ Get a reference to this module's variable namespace")
    print("-> variables = Dictionary(locals())")
    variables = Dictionary(globals())
    print()

    print("++ Get a reference to your filesystem")
    print("-> files = FileSystem()")
    files = FileSystem()
    print()

    print("++ Define example data")
    print('-> data = {"a": "Athabasca", "b": "Bosphorus", "c": "Cerulean"}')
    data = {"a": "Athabasca", "b": "Bosphorus", "c": "Cerulean"}
    print()

    print("++ Write example data to file as JSON")
    print('-> (variables["data"] >> files["output/data.json"])()')
    print('NOTE: This creates example.json in your working directory')
    (variables["data"] >> files["example.json"])()
    print()

    print("++ Read data back into the variable 'restored'")
    print('-> (variables["restored"] << files["example.json"])(target=dict)')
    (variables["restored"] << files["example.json"])(target=dict)
    print()

    print("++ Contents of the 'restored' variable")
    print("print(restored)")
    print(restored)
    print()

    print("++ Define some example tabular data in a Pandas DataFrame")
    print('-> example = DataFrame({"a": ["Athabasca"], "b": ["Bosphorus"], "c": ["Cerulean"]})')
    example = DataFrame({"a": ["Athabasca"], "b": ["Bosphorus"], "c": ["Cerulean"]})
    print()

    print("++ Get a reference to a SQLite table namespace")
    print('-> tables = Database("sqlite:///example.db")')
    print("++ NOTE: This creates example.db in your working directory")
    tables = Database("sqlite:///example.db")
    print()

    print("++ Populate an example table with the DataFrame")
    print('-> (variables["example"] >> tables["example"])()')
    (variables["example"] >> tables["example"])()
    print()

    print("++ Switch catalog to one that computes multi-step transfers")
    print("-> Transfer.catalog = PathTransfers()")
    Transfer.catalog = PathTransfers()
    print()

    print("++ Transfer data from a SQLite table to a Parquet file")
    print("   (internally, uses a DataFrame as an intermediate)")
    print('-> (tables["example"] >> files["example.parquet"])()')
    print("NOTE: This creates example.parquet in your working directory")
    (tables["example"] >> files["example.parquet"])()
    print()

    print("++ (Take a look at the contents of the 'example' table in example.db")
    print("    and the file example.parquet)")
    print()



"""
An implementation of a variable namespace.
"""

from typing import Any, Type
from uuid import uuid4

from vertebrate.data import (
    Namespace,
    Location,
    Address,
    Temporary,
    TemporaryLocation,
)


class Value(Location):

    def get(self) -> Any:
        return self.namespace.dict_[self.address]

    def set(self, val: Any) -> None:
        self.namespace.dict_[self.address] = val


class TemporaryValue(Value, TemporaryLocation):
    pass


class Dictionary(Namespace, Temporary):

    def __init__(self, dict_: dict[str, Any]):
        self.dict_ = dict_

    def type(self, address: Address) -> type:
        return type(self.dict_[address])

    @property
    def LocationType(self) -> Type[Location]:
        return Value

    def __contains__(self, address: Address) -> bool:
        return address in self.dict_

    def __delitem__(self, address: Address) -> None:
        del self.dict_[address]

    def temp(self) -> TemporaryLocation:
        temp_key = f"tmp_{str(uuid4()).replace('-', '_')}"
        return TemporaryValue(self, temp_key)



"""
An implementation of a table namespace.
"""

from contextlib import closing
from typing import Type

from sqlalchemy import create_engine, exc, MetaData, Table as SQLTable
from sqlalchemy.orm import Session

from vertebrate.data import Namespace, Location, Address


class Table(Location):

    def __call__(self):
        return SQLTable(
            self.address,
            self.namespace.meta,
            autoload_with=self.namespace.engine,
        )


class Database(Namespace):

    def __init__(self, db_url: str, *args, **kwargs) -> None:
        self.engine = create_engine(db_url, *args, **kwargs)
        self.meta = MetaData()

    def type(self, address: Address) -> type:
        # Rudimentary table typing
        return SQLTable

    @property
    def LocationType(self) -> Type[Location]:
        return Table

    def __contains__(self, address: Address) -> bool:
        try:
            table = self[address]()
            return True
        except exc.NoSuchTableError:
            return False

    def __delitem__(self, address: Address) -> None:
        self[address]().drop(self.engine)



"""
Example usage of the Environment base class using the OSEnvironment built-in.
"""

from vertebrate.builtins.os import OSEnvironment


def main():
    print("++ Get a reference to the OS environment -> local_machine = OSEnvironment()")
    local_machine = OSEnvironment()
    print()

    print('++ Reference the ls executable -> ls = local_machine["ls"]')
    ls = local_machine["ls"]
    print()

    print('++ Execute `ls -l .` -> ls("-l", ".")')
    print(local_machine["ls"]("-l", ".").decode("utf-8"))
    # ^^ Return type is bytes
    print()

    msg = "++ Environment variables can be passed in as kwargs"
    msg += '\n-> sh("-c", "printf $item", item="hello")'
    print(msg)
    print(local_machine["sh"]("-c", "printf $item", item="hello").decode("utf-8"))
    print()

    print('++ Is cd installed on your local machine? -> "cd" in local_machine')
    print("cd" in local_machine)
    print()

    print('++ Verify that this can return false -> "not_a_program" in local_machine')
    print("not_a_program" in local_machine)
    print()


if __name__ == "__main__":
    main()



"""
ABCs for data transfer.
"""

from abc import ABC, abstractmethod
from collections.abc import Callable
from types import TracebackType
from typing import Any, Type


type Address = str


class Namespace(ABC):
    """
    A typed collection of locations where data resides.
    """

    @abstractmethod
    def type(self, address: Address) -> type | None:
        pass

    @abstractmethod
    def __contains__(self, address: Address) -> bool:
        pass

    @abstractmethod
    def __delitem__(self, address: Address) -> None:
        pass

    @property
    @abstractmethod
    def LocationType(self) -> Type["Location"]:
        pass

    def __getitem__(self, address: Address) -> "Location":
        return self.LocationType(self, address)


class Temporary(ABC):
    """
    Mix-in for namespaces that implement temporary locations.
    """

    @abstractmethod
    def temp(self) -> "TemporaryLocation":
        pass


class Location(object):
    """
    A location that data can be read from and written to.
    The exact semantics of its read and write operations differ
    by implementation (ex. a simple file, a message queue, etc.).
    """

    def __init__(self, namespace: Namespace, address: Address) -> None:
        self._namespace = namespace
        self._address = address

    @property
    def namespace(self) -> Namespace:
        return self._namespace

    @property
    def address(self) -> Address:
        return self._address

    @property
    def type(self) -> type:
        return self.namespace.type(self.address)

    def __rshift__(self, loc: "Location") -> "Transfer":
        """
        A data transfer where this location is the source.
        """
        return Transfer(source=self, target=loc)

    def __lshift__(self, loc: "Location") -> "Transfer":
        """
        A data transfer where this location is the target.
        """
        return Transfer(source=loc, target=self)


class TemporaryLocation(Location):
    """
    A temporary location. Adds a context manager for location clean-up.
    """

    def __enter__(self) -> "TemporaryLocation":
        return self

    def __exit__(
        self,
        type_: type[BaseException] | None,
        value: BaseException | None,
        traceback: TracebackType | None,
    ) -> bool | None:
        del self.namespace[self.address]


type LocationTypes = tuple[type, type]
type TransferImplementation = Callable[[Location, Location], None]


class Transfers(ABC):
    """
    A collection of known transfers between datatypes
    (ex. dict to JSON, JSON to dict, database table to Parquet file, etc.).
    """

    @abstractmethod
    def __getitem__(self, datatypes: LocationTypes) -> TransferImplementation | None:
        pass

    @abstractmethod
    def __contains__(self, datatypes: LocationTypes) -> bool:
        pass


class Transfer(object):
    """
    A transfer of data from one location to another.
    """

    catalog: Transfers | None = None

    def __init__(self, source: Location, target: Location) -> None:
        self._source = source
        self._target = target

    @property
    def source(self) -> Location:
        return self._source

    @property
    def target(self) -> Location:
        return self._target

    def __call__(
        self,
        source: type | None = None,
        target: type | None = None,
        transfer: TransferImplementation | None = None,
    ) -> None:
        source_type = source if source else self.source.type
        target_type = target if target else self.target.type
        if not transfer:
            transfer = self.catalog[source_type, target_type]
        if not transfer:
            msg = f"No transfer defined from {source_type} to {target_type}."
            raise NotImplementedError(msg)
        transfer(self.source, self.target)



"""
A simple implementation of an order namespace.
"""

from typing import Any, Type

from vertebrate.builtins.python import PythonEnvironment
from vertebrate.builtins.os import OSEnvironment
from vertebrate.compute import Command, Environment
from vertebrate.data import Address
from vertebrate.plan import Order, Orders


class SimpleOrder(Order):

    @property
    def info(self) -> dict[str, Any]:
        return self.namespace.spec[self.address]

    @property
    def requirements(self):
        return self.info["requirements"]

    @property
    def command(self) -> Command:
        env = self.namespace.environment(self.info["env"])
        action = env[self.info["exe"]]
        return Command(action, args=self.info["args"])


class SimpleOrders(Orders):

    def __init__(self, spec: dict[str, Any], pyenv: PythonEnvironment) -> None:
        self.spec = spec
        self.pyenv = pyenv
        self.osenv = OSEnvironment()

    def type(self, address: Address) -> type | None:
        # No meaningful notion of order type for now
        if address in self:
            return Order

    def __contains__(self, address: Address) -> bool:
        return address in self.spec

    def __delitem__(self, address: Address) -> None:
        raise NotImplementedError("Deletion not implemented for SimpleOrders")

    @property
    def LocationType(self) -> Type[Order]:
        return SimpleOrder

    def PythonEnvironment(self) -> PythonEnvironment:
        return self.pyenv

    def OSEnvironment(self) -> OSEnvironment:
        return self.osenv

    def environment(self, env: str) -> Environment:
        return getattr(self, env)()



"""
Example usage of the Environment base class using the PythonEnvironment built-in.
"""

import sys

from vertebrate.builtins.python import PythonEnvironment


type Numeric = int | float


def subtract(a: Numeric, b: Numeric) -> Numeric:
    return a - b


class AddTo(object):

    def __init__(self, x: Numeric):
        self.x = x

    def plus(self, y: Numeric):
        return self.x + y


def main():
    print("++ Get a reference to this module")
    print("-> this_module = PythonEnvironment(sys.modules[__name__])")
    this_module = PythonEnvironment(sys.modules[__name__])
    print()

    print("++ Is there a function named 'subtract' in this module?")
    print('-> "subtract" in this_module')
    print("subtract" in this_module)
    print()

    print('++ Execute it -> this_module["subtract"](5.0, 3)')
    print(this_module["subtract"](5.0, 3))
    print()

    print("++ Construct a PythonEnvironment from an object")
    print("-> five = PythonEnvironment(AddTo(5))")
    five = PythonEnvironment(AddTo(5))
    print()

    print('++ Execute one of its methods -> five["plus"](4)')
    print(five["plus"](4))
    print()


if __name__ == "__main__":
    main()
